/*
 * weight-grader - ad7730.c
 *
 * Analog Devices AD7730 load cell userspace driver
 *
 * Copyright (C) 2017  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>
#include <pu-lib.h>

typedef __u8 u8;
typedef __u16 u16;
typedef __u32 u32;
typedef __s32 s32;

#define AD7730_STATUS		0
#define AD7730_STATUS_RDY	0x80
#define AD7730_DATA		1
#define AD7730_MODE		2
#define AD7730_MODE_IZEROCALIB	0x8000
#define AD7730_MODE_IFULLCALIB	0xa000
#define AD7730_MODE_CONTINOUS	0x2000
#define AD7730_MODE_UNIPOLAR	0x1000
#define AD7730_MODE_16BIT	0x0000
#define AD7730_MODE_24BIT	0x0100
#define AD7730_MODE_HIREF	0x0080
#define AD7730_MODE_RANGE_10	0x0000
#define AD7730_MODE_RANGE_20	0x0010
#define AD7730_MODE_RANGE_40	0x0020
#define AD7730_MODE_RANGE_80	0x0030

#define AD7730_FILTER		3
#define AD7730_FILTER_SF_SHIFT	12
#define AD7730_FILTER_SKIP	0x000200
#define AD7730_FILTER_FAST	0x000100
#define AD7730_FILTER_AC	0x000020
#define AD7730_FILTER_CHOP	0x000010
#define AD7730_DAC		4
#define AD7730_OFFSET		5
#define AD7730_GAIN		6
#define AD7730_TEST		7

static void dumpstat(const char *name, int fd)
{
	__u8	lsb, bits, mode;
	__u32	speed;

	if (ioctl(fd, SPI_IOC_RD_MODE, &mode) < 0) {
		perror("SPI rd_mode");
		return;
	}
	if (ioctl(fd, SPI_IOC_RD_LSB_FIRST, &lsb) < 0) {
		perror("SPI rd_lsb_fist");
		return;
	}
	if (ioctl(fd, SPI_IOC_RD_BITS_PER_WORD, &bits) < 0) {
		perror("SPI bits_per_word");
		return;
	}
	if (ioctl(fd, SPI_IOC_RD_MAX_SPEED_HZ, &speed) < 0) {
		perror("SPI max_speed_hz");
		return;
	}

	info("%s: spi mode 0x%x, %d bits %sper word, %d Hz max",
		name, mode, bits, lsb ? "(lsb first) " : "", speed);
}

int ad7730_readb(int fd, unsigned int reg)
{
	struct spi_ioc_transfer xfer;
	unsigned char rx[2];
	unsigned char tx[2];
	int ret;

	if (reg >= 8)
		return -EINVAL;

	memset(tx, 0, sizeof(tx));
	memset(rx, 0, sizeof(rx));
	tx[0] = 0x10 | reg;
	
	memset(&xfer, 0, sizeof(xfer));
	xfer.len = 2;
	xfer.tx_buf = (unsigned long) tx;
	xfer.rx_buf = (unsigned long) rx;

	ret = ioctl(fd, SPI_IOC_MESSAGE(1), &xfer);
	if (ret < 0) {
		perror("ioctl(SPI_IOC_MESSAGE)");
		exit(EXIT_FAILURE);
	}

	return rx[1];
}

int ad7730_readw(int fd, unsigned int reg)
{
	struct spi_ioc_transfer xfer;
	unsigned char rx[3];
	unsigned char tx[3];
	int ret;

	if (reg >= 8)
		return -EINVAL;

	memset(tx, 0, sizeof(tx));
	memset(rx, 0, sizeof(rx));
	tx[0] = 0x10 | reg;
	
	memset(&xfer, 0, sizeof(xfer));
	xfer.len = 3;
	xfer.tx_buf = (unsigned long) tx;
	xfer.rx_buf = (unsigned long) rx;

	ret = ioctl(fd, SPI_IOC_MESSAGE(1), &xfer);
	if (ret < 0) {
		perror("ioctl(SPI_IOC_MESSAGE)");
		exit(EXIT_FAILURE);
	}

	return (rx[1] << 8) | rx[2];
}

int ad7730_read24(int fd, unsigned int reg)
{
	struct spi_ioc_transfer xfer;
	unsigned char rx[4];
	unsigned char tx[4];
	int ret;

	if (reg >= 8)
		return -EINVAL;

	memset(tx, 0, sizeof(tx));
	memset(rx, 0, sizeof(rx));
	tx[0] = 0x10 | reg;
	
	memset(&xfer, 0, sizeof(xfer));
	xfer.len = 4;
	xfer.tx_buf = (unsigned long) tx;
	xfer.rx_buf = (unsigned long) rx;

	ret = ioctl(fd, SPI_IOC_MESSAGE(1), &xfer);
	if (ret < 0) {
		perror("ioctl(SPI_IOC_MESSAGE)");
		exit(EXIT_FAILURE);
	}

	return (rx[1] << 16) | (rx[2] << 8) | rx[3];
}

int ad7730_writew(int fd, u16 x, unsigned int reg)
{
	struct spi_ioc_transfer xfer;
	unsigned char rx[3];
	unsigned char tx[3];
	int ret;

	if (reg >= 8)
		return -EINVAL;

	memset(tx, 0, sizeof(tx));
	memset(rx, 0, sizeof(rx));
	tx[0] = 0x00 | reg;
	tx[1] = (x >> 8) & 0xff;
	tx[2] = (x >> 0) & 0xff;
	
	memset(&xfer, 0, sizeof(xfer));
	xfer.len = 3;
	xfer.tx_buf = (unsigned long) tx;
	xfer.rx_buf = (unsigned long) rx;

	ret = ioctl(fd, SPI_IOC_MESSAGE(1), &xfer);
	if (ret < 0) {
		perror("ioctl(SPI_IOC_MESSAGE)");
		exit(EXIT_FAILURE);
	}
	return 0;
}

int ad7730_write24(int fd, unsigned int x, unsigned int reg)
{
	struct spi_ioc_transfer xfer;
	unsigned char rx[4];
	unsigned char tx[4];
	int ret;

	if (reg >= 8)
		return -EINVAL;

	memset(tx, 0, sizeof(tx));
	memset(rx, 0, sizeof(rx));
	tx[0] = 0x00 | reg;
	tx[1] = (x >> 16) & 0xff;
	tx[2] = (x >> 8) & 0xff;
	tx[3] = (x >> 0) & 0xff;
	
	memset(&xfer, 0, sizeof(xfer));
	xfer.len = 4;
	xfer.tx_buf = (unsigned long) tx;
	xfer.rx_buf = (unsigned long) rx;

	ret = ioctl(fd, SPI_IOC_MESSAGE(1), &xfer);
	if (ret < 0) {
		perror("ioctl(SPI_IOC_MESSAGE)");
		exit(EXIT_FAILURE);
	}
	return 0;
}

int ad7730_init(const char *spi_dev, unsigned int freq)
{
	u8 mode = SPI_MODE_1;
	u32 speed = 500000;
	int ret;
	int fd;
	int gpio_fd;
	int x;
	int gpio;
	s32 sum = 0;
	unsigned int i;
	unsigned int sf_freq;

	fd = open(spi_dev, O_RDWR);
	if (fd == -1) {
		perror("open");
		exit(EXIT_FAILURE);
	}

	if (ioctl(fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed) < 0) {
		perror("SPI max_speed_hz");
		return;
	}

	ret = ioctl(fd, SPI_IOC_WR_MODE, &mode);
	if (ret < 0) {
		perror("ioctl(SPI_IOC_WR_MODE)");
		exit(EXIT_FAILURE);
	}

	dumpstat("ad7730", fd);

	x = ad7730_readb(fd, AD7730_STATUS);
	info("STATUS: %02x", x);

	x = ad7730_readw(fd, AD7730_MODE);
	info("MODE: %04x", x);
	
	x = ad7730_read24(fd, AD7730_FILTER);
	info("FILTER: %06x", x);
	
	x = ad7730_readb(fd, AD7730_DAC);
	info("DAC: %02x", x);

	x = ad7730_read24(fd, AD7730_OFFSET);
	info("OFFSET: %06x", x);

	x = ad7730_read24(fd, AD7730_GAIN);
	info("GAIN: %06x", x);
	
//	x = (2047 << AD7730_FILTER_SF_SHIFT);
//	x = (256 << AD7730_FILTER_SF_SHIFT);
//	x = (512 << AD7730_FILTER_SF_SHIFT) | AD7730_FILTER_SKIP;
//	x = (512 << AD7730_FILTER_SF_SHIFT) | AD7730_FILTER_CHOP | 15;
	x = (2048 << AD7730_FILTER_SF_SHIFT)
		| AD7730_FILTER_SKIP
		| AD7730_FILTER_AC
		| AD7730_FILTER_CHOP | 15;

	ad7730_write24(fd, x, AD7730_FILTER);
	x = ad7730_read24(fd, AD7730_FILTER);
	info("FILTER: %04x", x);

#if 1
	fprintf(stderr, "Full-scale calibration...");
	x = AD7730_MODE_IFULLCALIB
		| AD7730_MODE_24BIT | AD7730_MODE_HIREF
		| AD7730_MODE_RANGE_10;
	ad7730_writew(fd, x, AD7730_MODE);
	do {
		x = ad7730_readb(fd, AD7730_STATUS);
		usleep(1000);
	} while (x & AD7730_STATUS_RDY);
	fprintf(stderr, "OK\n");
	
	fprintf(stderr, "Zero-scale calibration...");
	x = AD7730_MODE_IZEROCALIB
		| AD7730_MODE_24BIT | AD7730_MODE_HIREF
		| AD7730_MODE_RANGE_10;
	ad7730_writew(fd, x, AD7730_MODE);
	do {
		x = ad7730_readb(fd, AD7730_STATUS);
		usleep(1000);
	} while (x & AD7730_STATUS_RDY);
	fprintf(stderr, "OK\n");
#endif

	x = ad7730_read24(fd, AD7730_OFFSET);
	info("OFFSET: %06x", x);

	x = ad7730_read24(fd, AD7730_GAIN);
	info("GAIN: %06x", x);

	x = AD7730_MODE_CONTINOUS
		| AD7730_MODE_24BIT | AD7730_MODE_HIREF
		| AD7730_MODE_RANGE_10;

	ad7730_writew(fd, x, AD7730_MODE);
	x = ad7730_readw(fd, AD7730_MODE);
	info("MODE: %04x", x);

	sf_freq = (102400 + freq/2) / freq;
	if (sf_freq < 75)
		sf_freq = 75;
	else if (sf_freq > 2048)
		sf_freq = 2048;

	info("SF = %d", sf_freq);

	x = (sf_freq << AD7730_FILTER_SF_SHIFT)
		| AD7730_FILTER_AC
		| AD7730_FILTER_CHOP | 15;
	ad7730_write24(fd, x, AD7730_FILTER);
	x = ad7730_read24(fd, AD7730_FILTER);
	info("FILTER: %04x", x);

	i = 0;
	do {
		x = ad7730_readb(fd, AD7730_STATUS);
		if (!(x & AD7730_STATUS_RDY)) {
			ad7730_read24(fd, AD7730_DATA) << 8;
			i++;
		}
	} while (i < 128);
	return fd;
}

int __load_cell_read(int fd, s32 *data)
{
	s32 x;

	x = ad7730_read24(fd, AD7730_DATA) << 8;
	x >>= 8;
	x += 1 << 23;
	*data = x;
	return 0;
}

int load_cell_read(int fd, s32 *data)
{
	unsigned int x;

	x = ad7730_readb(fd, AD7730_STATUS);
	if (!(x & AD7730_STATUS_RDY))
		return __load_cell_read(fd, data);
	return -EAGAIN;
}
