/*
 * weight-grader - wg-graber.c
 *
 * Copyright (C) 2017  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pu-lib.h>
#include <pu-buff.h>
#include <pu-proto.h>
#include <pu-proto-lib.h>
#include <pugcp-proto.h>
#include <pugcp-proto-lib.h>
#include <gcp.h>
#include <list.h>
#include <time.h>
#include <poll.h>
#include <mclock.h>
#include <sched.h>
#include <ad7730.h>
#include <xbat.h>

#ifndef PU_MTU
#define PU_MTU		1048576
#endif

struct pu_sock {
	int fd;
	int datagram;
};

static const char *server_host = "localhost";
static const char *server_port = "10000";
static const char *config_file;
static const char *spi_dev = "/dev/spidev1.0";
static const char *xbat_file;
static int rdy_pin = 60;
static int verbose;
static unsigned int freq = 400;

static int xbat_fd = -1;

static const char copyright_string[] =
"wg-graber " PACKAGE_VERSION ".\n"
"Copyright (C) 2017  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_copyright(void)
{
	puts(copyright_string);
}

static const char usage[] =
"usage: %s [options] [--] <id...>\n"
"Options:\n"
"  -V, --version              show version\n"
"  -h, --help                 show this help\n"
"  -p, --port=PORT            use specified port (3000)\n"
"  -s, --host=HOST            use specified target host (localhost)\n"
"\n"
"Report bugs to krzysiek@podlesie.net.\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, name ? name : "wg-graber");
	exit(EXIT_SUCCESS);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **argv)
{
	static struct option long_options[] = {
		{ "config", 1, NULL, 'c' },
		{ "spi", 1, NULL, 'S' },
		{ "rdy-pin", 1, NULL, 'R' },
		{ "freq", 1, NULL, 'f' },
		{ "help", 0, NULL, 'h' },
		{ "host", 1, NULL, 's' },
		{ "port", 1, NULL, 'p' },
		{ "version", 0, NULL, 'V' },
		{ "xbat", 1, NULL, 'x' },
		{ NULL, 0, NULL, 0 },
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, argv, "c:df:hpR::s:S:Vx:",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'c':
			xfree(config_file);
			config_file = strdup(config_file);
			break;
		case 'f':
			freq = atoi(optarg);
			break;
		case 'h':
			show_usage(argv[0]);
			BUG();
		case 'p':
			server_port = strdup(optarg);
			break;
		case 'R':
			rdy_pin = atoi(optarg);
			break;
		case 'S':
			xfree(spi_dev);
			spi_dev = strdup(optarg);
			break;
		case 's':
			server_host = strdup(optarg);
			break;
		case 'V':
			show_version();
			BUG();
		case 'x':
			xfree(xbat_file);
			xbat_file = strdup(optarg);
			break;
		default:
			die("unknown option: %d.", c);
		}
	}
}

static int do_connect(const char *host, const char *port, int datagram)
{
	struct addrinfo hints, *ai, *aitop;
	int ret;
	int fd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	if (datagram)
		hints.ai_socktype = SOCK_DGRAM;
	hints.ai_flags = 0;
	ret = getaddrinfo(host, port, &hints, &aitop);
	if (ret < 0) {
		warn("getaddrinfo(%s, %s): %s\n", host, port, gai_strerror(errno));
		return -errno;
	}

	for (ai = aitop; ai; ai = ai->ai_next) {
		fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (fd == -1) {
			warn("socket: %s\n", strerror(errno));
			continue;
		}

		ret = connect(fd, ai->ai_addr, ai->ai_addrlen);
		if (ret < 0) {
			warn("connect: %s\n", strerror(errno));
			close(fd);
			continue;
		}
		freeaddrinfo(aitop);
		return fd;
	}
	warn("can't connect to %s:%s", host, port);
	return -EAGAIN;
}

int pu_proto_xmit(void *priv, struct pu_buff *skb)
{
	int fd = *((int *) priv);

	write_or_die(fd, skb->data, skb->len);
	pu_consume_skb(skb);
	return 0;
}

static struct pu_buff *pu_proto_recv_datagram(void *priv)
{
	struct pu_header header;
	struct pu_buff *skb;
	struct pu_header *hdr;
	unsigned char *p;
	int fd = *((int *) priv);
	unsigned long len;
	ssize_t rd;

	rd = xrecv(fd, &header, sizeof(header), MSG_PEEK);
	if (rd != sizeof(header)) {
		if (errno == EAGAIN)
			return NULL;
		die_errno("cannot read header");
	}

	if (header.magic != PU_MAGIC)
		die("protocol error: invalid magic %02x", header.magic);

	len = be32_to_cpu(header.data_bytes);
	if (len & PU_DATA_ERROR)
		len = 0;
	len = pu_align_data(len);
	len += sizeof(*hdr);

	skb = pu_alloc_skb(len, GFP_KERNEL);
	p = pu_skb_put(skb, 0);
	rd = xrecv(fd, p, len, 0);
	if (rd != len)
		die_errno("cannot read packet");

	hdr = (void *) skb->data;
	if (hdr->magic != PU_MAGIC)
		die("protocol error: invalid magic %02x", hdr->magic);

	len = be32_to_cpu(hdr->data_bytes);
	if ((len > 0) && !(len & PU_DATA_ERROR))
		pu_skb_trim(skb, sizeof(*hdr) + len);
	pu_dump_skb(skb);
	return skb;
}

static struct pu_buff *pu_proto_recv_stream(void *priv)
{
	struct pu_buff *skb;
	struct pu_header *hdr;
	int fd = *((int *) priv);
	unsigned long len;
	ssize_t rd;

	skb = pu_alloc_skb(512, GFP_KERNEL);
	rd = read_in_full(fd, pu_skb_put(skb, sizeof(*hdr)), sizeof(*hdr));
	if (rd != sizeof(*hdr))
		die_errno("cannot read header");

	hdr = (void *) skb->data;
	if (hdr->magic != PU_MAGIC)
		die("protocol error: invalid magic %02x", hdr->magic);

	len = be32_to_cpu(hdr->data_bytes);
	if ((len > 0) && !(len & PU_DATA_ERROR)) {
		unsigned int align = pu_align_data(len);
		struct pu_buff *skb2;

		if (align >= PU_MTU)
			die("message to large: %d bytes", len);

		skb2 = skb_copy_expand(skb, 0, align, GFP_KERNEL);
		pu_consume_skb(skb);
		skb = skb2;

		rd = read_in_full(fd, pu_skb_put(skb, align), align);
		if (rd != align)
			die_errno("cannot read payload");

		pu_skb_trim(skb, sizeof(*hdr) + len);
	}
	return skb;
}

static struct pu_buff *pu_proto_recv(void *priv)
{
	struct pu_sock *sock = priv;

	if (sock->datagram)
		return pu_proto_recv_datagram(priv);
	return pu_proto_recv_stream(priv);
}

static struct pu_buff *pugcp_proto_recv(void *priv, unsigned int *seq,
		unsigned int *node, unsigned int *command)
{
	struct pu_buff *skb;
	struct pu_header *hdr;
	struct pugcp_header *pugcp;

	skb = pu_proto_recv(priv);
	if (!skb)
		return skb;
	hdr = (void *) skb->data;
	pu_skb_pull(skb, sizeof(*hdr));
	if (hdr->proto != PU_PROTO_GRADER) {
		warn("received frame with unsupported PU protocol: 0x%02x\n",
				be16_to_cpu(hdr->proto));
		pu_free_skb(skb);
		return NULL;
	}
	pugcp = (void *) skb->data;
	pu_skb_pull(skb, sizeof(*pugcp));
	if (be32_to_cpu(pugcp->proto) != GCP_PROTO_SEQCMD) {
		warn("received frame with unsupported GCP protocol: 0x%02x\n",
				be32_to_cpu(hdr->proto));
		pu_free_skb(skb);
		return NULL;
	}
	*seq = be16_to_cpu(hdr->id);
	*node = be32_to_cpu(pugcp->node);
	*command = be32_to_cpu(pugcp->command);
	return skb;
}

static void *pu_proto_verify(struct pu_buff *skb, unsigned int proto,
		unsigned int id)
{
	struct pu_header *hdr = (void *) skb->data;
	unsigned int pkt_id;
	unsigned int len;

	if (hdr->proto != proto) {
		die("protocol error: invalid subprotocol: %d, expected %d",
				hdr->proto, proto);
	}

	pkt_id = be16_to_cpu(hdr->id);
	if (pkt_id != id) {
		die("protocol error: invalid id: %d, expected %d",
				pkt_id, id);
	}

	len = be32_to_cpu(hdr->data_bytes);
	if (len & PU_DATA_ERROR) {
		warn("protocol warning: id: %d, remote error: %d",
				pkt_id, len & ~PU_DATA_ERROR);
		return NULL;
	}

	return pu_skb_pull(skb, sizeof(*hdr));
}

static void *pugcp_proto_verify(struct pu_buff *skb, unsigned int proto,
		unsigned int node, unsigned int command,
		unsigned int id)
{
	struct pugcp_header *hdr = pu_proto_verify(skb, PU_PROTO_GRADER, id);
	unsigned int pkt_proto;
	unsigned int pkt_node;
	unsigned int pkt_command;

	if (!hdr)
		return NULL;

	pkt_proto = be32_to_cpu(hdr->proto);
	if (pkt_proto != proto) {
		die("protocol error: invalid GCP protocol: %d, expected %d",
				hdr->proto, proto);
	}

	pkt_node = be32_to_cpu(hdr->node);
	if (pkt_node != node) {
		die("protocol error: invalid node: %d, expected %d",
				pkt_node, node);
	}

	pkt_command = be32_to_cpu(hdr->command);
	if (pkt_command != command) {
		die("protocol error: invalid node: %d, expected %d",
				pkt_command, command);
	}
	return pu_skb_pull(skb, sizeof(*hdr));
}

#if 0
static u32 grader_psched_read(struct pu_sock *sock, unsigned int node,
		unsigned int command, unsigned int id)
{
	struct pu_buff *skb;
	struct pugcp_psched *p;
	struct pugcp_psched *resp;
	u32 ret;

	skb = pugcp_proto_alloc_skb(sizeof(*p));
	p = (void *) pu_skb_put(skb, sizeof(*p));
	if (id > GCP_PSCHED_ID_MASK)
		die("too large id '%d'\n", id);
	p->id = cpu_to_be32(id);
	p->value = cpu_to_be32(0);

	pugcp_proto_xmit_new(sock, skb, 0, node, command, 0);

	skb = pu_proto_recv(sock);
	resp = pugcp_proto_verify(skb, 0, node, command, 0);
	if (!resp)
		die("read error");

	ret = be32_to_cpu(resp->value);
	pu_consume_skb(skb);
	return ret;
}
#endif

static void grader_psched_write(struct pu_sock *sock, unsigned int node,
		unsigned int command, unsigned int id, unsigned int value)
{
	struct pu_buff *skb;
	struct pugcp_psched *p;
	struct pugcp_psched *resp;

	skb = pugcp_proto_alloc_skb(sizeof(*p));
	p = (void *) pu_skb_put(skb, sizeof(*p));
	if (id > GCP_PSCHED_ID_MASK)
		die("too large id '%d'\n", id);
	p->id = cpu_to_be32(0x80000000 | id);
	p->value = cpu_to_be32(value);
	
	pugcp_proto_xmit_new(sock, skb, 0, node, command, 0);

	skb = pu_proto_recv(sock);
	if (!skb)
		return;
	resp = pugcp_proto_verify(skb, 0, node, command, 0);
	if (!resp)
		die("write error");
	if (be32_to_cpu(resp->id) != id)
		die("invalid returned id");
	if (be32_to_cpu(resp->value) != value)
		die("invalid returned value: written %d, read %d\n",
				value, be32_to_cpu(resp->value));
	pu_consume_skb(skb);
}

static void grader_cadd_event(struct pu_sock *sock, unsigned int node,
		unsigned int pos, unsigned int output)
{
	struct pu_buff *skb;
	struct pugcp_cadd_event *p;

	skb = pugcp_proto_alloc_skb(sizeof(*p));
	p = (void *) pu_skb_put(skb, sizeof(*p));
	p->output = cpu_to_be32(output);
	p->position = cpu_to_be32(pos);

	pugcp_proto_xmit_new(sock, skb, 0, node, GCP_CADD_EVENT, 0);
}

static void grader_enable_event_info(struct pu_sock *sock, unsigned int node,
		unsigned int mask, u64 cookie)
{
	struct pu_buff *skb;
	struct pugcp_cpevent_enable_info *p;

	skb = pugcp_proto_alloc_skb(sizeof(*p));
	p = (void *) pu_skb_put(skb, sizeof(*p));
	p->mask = cpu_to_be32(mask);
	p->reserved = cpu_to_be32(0);
	p->cookie = cookie;

	pugcp_proto_xmit_new(sock, skb, 0, node, PUGCP_CPEVENT_ENABLE_INFO, 0);

#if 0
	struct pugcp_cpevent_enable_info *resp;
	skb = pu_proto_recv(sock);
	resp = pugcp_proto_verify(skb, 0, node, PUGCP_CPEVENT_ENABLE_INFO, 0);
	if (!resp)
		die("write error");
	if (be32_to_cpu(resp->mask) != 1)
		die("invalid returned mask");
	if (resp->cookie != cookie)
		die("invalid returned cookie: written %016lld, read %016llx\n",
				cookie, resp->cookie);
	pu_consume_skb(skb);
#endif
}

static void pu_close(struct pu_sock *s)
{
	int warned = 0;
	ssize_t rd;
	char c;

	if (!s->datagram) {
		shutdown(s->fd, SHUT_WR);
		do {
			rd = xread(s->fd, &c, sizeof(c));
			if (rd < 0)
				die_errno("read error");
			if (rd && !warned) {
				warn("received garbage\n");
				warned = 1;
			}
		} while (rd);
	}
	close(s->fd);
}

struct line_state {
	unsigned int seq;
	int weight;
};

static void pugcp_handle_cpevent_info(struct pu_sock *sock, struct pu_buff *skb,
		unsigned int seq, struct line_state *state)
{
	struct pugcp_cpevent_info *i = (void *) skb->data;
	unsigned int position;
	unsigned int events;

	if (skb->len < offsetof(struct pugcp_cpevent_info, time_sec)) {
		warn("too short CPEVENT_INFO packet: %d bytes\n", skb->len);
		return;
	}
	position = be32_to_cpu(i->position);
	events = be32_to_cpu(i->output);

	if (verbose >= 0) {
		info("event %d: events %08x, position: %08x",
				seq, events, position);
	}

	if (state->seq == seq) {
		int output;

		printf("wg-v0 0x%08x %d\n", position, state->weight);
	} else if (state->seq != -1) {
		warn("missed CPEVENT_INFO, is %d, should be %d",
				seq, state->seq);
	}
	state->seq = (seq + 1) & 0xffff;
}

static void pugcp_handle_command(struct pu_sock *sock,
		struct line_state *state)
{
	struct pu_buff *skb;
	unsigned int command;
	unsigned int node;
	unsigned int seq;

	skb = pugcp_proto_recv(sock, &seq, &node, &command);
	if (!skb)
		return;

	if (node != 14)
		return;

	switch (command) {
	case GCP_CPEVENT_INFO:
		pugcp_handle_cpevent_info(sock, skb, seq, state);
		break;
	case PUGCP_CPEVENT_ENABLE_INFO:
		pu_consume_skb(skb);
		break;
	default:
		warn("unknown command: 0x%03x\n", command);
		pu_consume_skb(skb);
	}
}

static int ad7730_fd;
static int rdy_fd;

struct rambat_record {
	u8 magic;		/* DATA_RECORD_MAGIC */
	u8 b_len;		/* number of binary words */
	u8 a_len;		/* number of analog inputs */
	u8 tcn;			/* time change notification */
	u64 timestamp;		/* timestamp, in nanoseconds */
	u16 data[0];		/* data */
} __attribute__ ((packed));

#define RAMBAT_RECORD_MAGIC	0xfb

static void xbat_push(const s16 *buf, unsigned int channels)
{
	unsigned int i, j;
	struct xbat_header *msg;
	struct rambat_record *rambat;
	int16_t *msg_data;
	unsigned int len;
	struct timespec ts;

	if (channels >= 256) {
		static int warned = 0;

		if (!warned) {
			warned = 1;
			warn("xbat output supports only 255 channels");
		}
		channels = 255;
	}
	clock_gettime(CLOCK_REALTIME, &ts);
	len = sizeof(*msg) + sizeof(*rambat) + channels * sizeof(*msg_data) + 2;
	msg = malloc(xbat_align(len));
	msg->version = 0;
	msg->proto = XBAT_PROTO_RAMBAT_CRC16;
	msg->length = cpu_to_be16(len);
	rambat = (void *)(msg + 1);
	rambat->magic = RAMBAT_RECORD_MAGIC;
	rambat->b_len = 0;
	rambat->a_len = channels;
	rambat->tcn = 0;
	rambat->timestamp = ts.tv_sec * 1000000000ULL + ts.tv_nsec;

	msg_data = (void *)(rambat + 1);
	memcpy(msg_data, buf, 2 * channels);
	msg_data[channels] = crc16(0xffff, (void *) rambat,
			len - 2 - sizeof(*msg));
	xwrite(xbat_fd, msg, xbat_align(len));
	free(msg);
}

static void main_loop(struct pu_sock *sock)
{
	struct timespec t;
	struct pollfd pfd[2] = { 0 };
	struct line_state state = { -1 };
	u64 cookie = 0;
	int ret;
	time_t next_enable;
	s32 w;
	s32 w1 = 0;
	s32 w2 = 0;
	unsigned int i = 0;
	char buf[16];

	pfd[0].fd = sock->fd;
	pfd[0].events = POLLIN;
	pfd[1].fd = rdy_fd;
	pfd[1].events = POLLPRI | POLLERR;
	ret = clock_gettime(CLOCK_MONOTONIC, &t);
	if (unlikely(ret)) {
		die("clock_gettime(CLOCK_MONOTONIC) error: %s",
				strerror(errno));
	}
	next_enable = t.tv_sec;

	for (;;) {
		ret = clock_gettime(CLOCK_MONOTONIC, &t);
		if (unlikely(ret)) {
			die("clock_gettime(CLOCK_MONOTONIC) error: %s",
					strerror(errno));
		}

		if (t.tv_sec >= next_enable) {
			grader_enable_event_info(sock, 14, 1, cookie);
			next_enable += 30;
		}

		lseek(rdy_fd, 0, SEEK_SET);
		ret = read(rdy_fd, buf, 16);
		if (ret < 1)
			die("gpio value read failure\n");
		ret = -EAGAIN;
		if (buf[0] == '1')
			ret = __load_cell_read(ad7730_fd, &w);
		w -= 176000;
//		w -= 254000;
//		w -= 254000;
		if (!ret) {
			int y;
			s16 wx[2];

#if 1
			w1 += w;
			y = w1 >> 2;
			w1 -= y;
			
			w2 += y;
			y = w2 >> 2;
			w2 -= y;
#else
			w1 += w;
			y = w1 >> 7;
			w1 -= y;
			
			w2 += y;
			y = w2 >> 7;
			w2 -= y;
#endif
			wx[0] = w >> 4;
			wx[1] = y >> 4;
			xbat_push(wx, 2);

			state.weight = y;
//			printf("%d	%d	%d\n", w, y, state.weight);
	
#if 0
			i++;
			if (i >= 400) {
				i = 0;
				printf("%.2f\n", y / 821.0);
			}
#endif
		}

		ret = poll(pfd, 2, 1000);
		if (ret < 0) {
			if (ret == EAGAIN || ret == EINTR)
				continue;
			warn("poll error: %s", strerror(errno));
			break;
		}
		if (!ret)
			continue;
		if (pfd[0].revents & (POLLERR | POLLHUP | POLLNVAL))
			break;
		if (pfd[0].revents & POLLIN)
			pugcp_handle_command(sock, &state);
	}
}

int main(int argc, char **argv)
{
	struct sched_param sched_param = {
		.sched_priority = 10,
	};
	struct pu_sock sock = { 0 };
	char path[PATH_MAX];

	parse_command_line(argc, argv);

	snprintf(path, PATH_MAX, "/sys/class/gpio/gpio%d/value", rdy_pin);
	rdy_fd = open(path, O_RDONLY);
	if (rdy_fd == -1) {
		die("can't open '%s': %s", path, strerror(errno));
	}

	if (xbat_file) {
		xbat_fd = open(xbat_file, O_RDWR | O_NOCTTY | O_NONBLOCK);
		if (fcntl(xbat_fd, F_SETFL, O_NONBLOCK))
			warn("fcntl(): %s", strerror(errno));
	}

	sched_setscheduler(0, SCHED_FIFO, &sched_param);

	mclock_init();
	ad7730_fd = ad7730_init(spi_dev, freq);
	for (;;) {
		sock.fd = do_connect(server_host, server_port, 1);
		if (sock.fd < 0) {
			sleep(1);
			continue;
		}
		sock.datagram = 1;
		if (fcntl(sock.fd, F_SETFL, O_NONBLOCK))
			warn("fcntl(): %s", strerror(errno));
		main_loop(&sock);
		pu_close(&sock);
	}
	return 0;
}
